@@ -20,12 +20,19 @@ static void
 test_bson_as_json (void)
 {
    bson_oid_t oid;
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   bson_decimal128_t decimal128;
+#endif
    bson_t *b;
    bson_t *b2;
    char *str;
    size_t len;
    int i;
 
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   decimal128.high = 0x3040000000000000ULL;
+   decimal128.low  = 0x000000000000000B;
+#endif
    bson_oid_init_from_string(&oid, "123412341234abcdabcdabcd");
 
    b = bson_new();
@@ -45,6 +52,9 @@ test_bson_as_json (void)
    assert(bson_append_minkey(b, "minkey", -1));
    assert(bson_append_maxkey(b, "maxkey", -1));
    assert(bson_append_symbol(b, "symbol", -1, "var a = {};", -1));
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   assert(bson_append_decimal128(b, "decimal128", -1, &decimal128));
+#endif
 
    b2 = bson_new();
    assert(bson_append_int32(b2, "0", -1, 60));
@@ -144,6 +154,31 @@ test_bson_as_json_double (void)
 }
 
 
+#ifdef BSON_EXPERIMENTAL_FEATURES
+static void
+test_bson_as_json_decimal128 (void)
+{
+   size_t len;
+   bson_t *b;
+   char *str;
+   bson_decimal128_t decimal128;
+   decimal128.high = 0x3040000000000000ULL;
+   decimal128.low  = 0x000000000000000B;
+
+   b = bson_new ();
+   assert (bson_append_decimal128 (b, "decimal128", -1, &decimal128));
+   str = bson_as_json (b, &len);
+   ASSERT_CMPSTR (str,
+                  "{ "
+                     "\"decimal128\" : { \"$numberDecimal\" : \"11\" }"
+                  " }");
+
+   bson_free(str);
+   bson_destroy(b);
+}
+#endif
+
+
 static void
 test_bson_as_json_utf8 (void)
 {
@@ -188,7 +223,7 @@ test_bson_as_json_stack_overflow (void)
    BSON_ASSERT(str);
 
    r = !!strstr(str, "...");
-   BSON_ASSERT(str);
+   BSON_ASSERT(r);
 
    bson_free(str);
    bson_destroy(&b);
@@ -252,7 +287,6 @@ test_bson_corrupt_utf8 (void)
    bson_free(buf);
 }
 
-
 static void
 test_bson_corrupt_binary (void)
 {
@@ -399,6 +433,22 @@ test_bson_json_read(void)
    _test_bson_json_read_compare(json, 5, first, second, third, NULL);
 }
 
+
+#ifdef BSON_EXPERIMENTAL_FEATURES
+static void
+test_bson_json_read_decimal128(void)
+{
+   const char * json = "{ \"decimal\" : { \"$numberDecimal\" : \"123.5\" }}";
+   bson_decimal128_t dec;
+   bson_t *doc;
+
+   bson_decimal128_from_string("123.5", &dec);
+   doc = BCON_NEW("decimal", BCON_DECIMAL128(&dec));
+
+   _test_bson_json_read_compare(json, 5, doc, NULL);
+}
+#endif
+
 static void
 test_json_reader_new_from_file (void)
 {
@@ -473,6 +523,22 @@ test_bson_json_read_missing_complex(void)
                          BSON_JSON_ERROR_READ_INVALID_PARAM);
 }
 
+static void
+test_bson_json_read_invalid_binary(void)
+{
+   bson_error_t error;
+   const char *json = "{ "
+       " \"bin\" : { \"$binary\" : \"invalid\", \"$type\" : \"80\" } }";
+   bson_t b;
+   bool r;
+   char *str;
+
+   r = bson_init_from_json (&b, json, -1, &error);
+   assert (!r);
+
+   bson_destroy (&b);
+}
+
 static void
 test_bson_json_read_invalid_json(void)
 {
@@ -554,6 +620,27 @@ test_bson_json_read_invalid(void)
    bson_destroy (&bson);
 }
 
+static void
+test_bson_json_binary_order (void)
+{
+   bson_error_t error;
+   const char *json = "{ \"bin\" : "
+       "{ \"$binary\" : \"IG43GK8JL9HRL4DK53HMrA==\", \"$type\" : \"05\" } }";
+   bson_t b;
+   bool r;
+   char *str;
+
+   r = bson_init_from_json (&b, json, -1, &error);
+   if (!r) fprintf (stderr, "%s\n", error.message);
+   assert (r);
+
+   str = bson_as_json (&b, NULL);
+   ASSERT_CMPSTR (str, json);
+
+   bson_destroy (&b);
+   bson_free (str);
+}
+
 static void
 test_bson_json_number_long (void)
 {
@@ -576,6 +663,25 @@ test_bson_json_number_long (void)
    assert (!bson_init_from_json (&b, json2, -1, &error));
 }
 
+static void
+test_bson_json_number_long_zero (void)
+{
+   bson_error_t error;
+   bson_iter_t iter;
+   const char *json = "{ \"key\": { \"$numberLong\": \"0\" }}";
+   bson_t b;
+   bool r;
+
+   r = bson_init_from_json (&b, json, -1, &error);
+   if (!r) fprintf (stderr, "%s\n", error.message);
+   assert (r);
+   assert (bson_iter_init (&iter, &b));
+   assert (bson_iter_find (&iter, "key"));
+   assert (BSON_ITER_HOLDS_INT64 (&iter));
+   assert (bson_iter_int64 (&iter) == 0);
+   bson_destroy (&b);
+}
+
 static const bson_oid_t *
 oid_zero (void)
 {
@@ -668,6 +774,29 @@ test_bson_json_dbref (void)
    }
 }
 
+#ifdef BSON_EXPERIMENTAL_FEATURES
+static void
+test_bson_json_number_decimal (void) {
+   bson_error_t error;
+   bson_iter_t iter;
+   bson_decimal128_t decimal128;
+   const char *json = "{ \"key\" : { \"$numberDecimal\": \"11\" }}";
+   bson_t b;
+   bool r;
+
+   r = bson_init_from_json (&b, json, -1, &error);
+   if (!r) fprintf (stderr, "%s\n", error.message);
+   assert(r);
+   assert (bson_iter_init (&iter, &b));
+   assert (bson_iter_find (&iter, "key"));
+   assert (BSON_ITER_HOLDS_DECIMAL128 (&iter));
+   bson_iter_decimal128 (&iter, &decimal128);
+   assert (decimal128.low == 11);
+   assert (decimal128.high == 0x3040000000000000ULL);
+   bson_destroy (&b);
+}
+#endif
+
 static void
 test_bson_json_inc (void)
 {
@@ -763,6 +892,28 @@ test_bson_json_date (void)
                               0);
 }
 
+static void
+test_bson_json_special_keys_at_top (void)
+{
+   const char *invalid[] = {
+      /* needs outer key name, like {x: {$date: {$numberLong: "1234"}}} */
+      "{ \"$date\" : { \"$numberLong\" : \"1234\" } }",
+      "{  \"$oid\" : { \"$numberLong\" : \"1234\" } }",
+      "{  \"$oid\" : \"563b029096f9bc35c61df3a3\" }",
+   };
+
+   bson_t b;
+   bson_error_t error;
+   int i;
+
+   for (i = 0; i < sizeof (invalid) / sizeof (const char *); i++) {
+      assert (!bson_init_from_json (&b, invalid[i], -1, &error));
+      ASSERT_CMPINT (error.domain, ==, BSON_ERROR_JSON);
+      ASSERT_CMPINT (error.code, ==, BSON_JSON_ERROR_READ_CORRUPT_JS);
+      ASSERT_CMPSTR (error.message, "Invalid MongoDB extended JSON");
+   }
+}
+
 static void
 test_bson_array_as_json (void)
 {
@@ -819,20 +970,30 @@ test_json_install (TestSuite *suite)
    TestSuite_Add (suite, "/bson/as_json/corrupt", test_bson_corrupt);
    TestSuite_Add (suite, "/bson/as_json/corrupt_utf8", test_bson_corrupt_utf8);
    TestSuite_Add (suite, "/bson/as_json/corrupt_binary", test_bson_corrupt_binary);
+   TestSuite_Add (suite, "/bson/as_json/binary_order", test_bson_json_binary_order);
    TestSuite_Add (suite, "/bson/as_json_spacing", test_bson_as_json_spacing);
+   TestSuite_Add (suite, "/bson/as_json_special_keys_at_top", test_bson_json_special_keys_at_top);
    TestSuite_Add (suite, "/bson/array_as_json", test_bson_array_as_json);
    TestSuite_Add (suite, "/bson/json/read", test_bson_json_read);
    TestSuite_Add (suite, "/bson/json/inc", test_bson_json_inc);
    TestSuite_Add (suite, "/bson/json/array", test_bson_json_array);
    TestSuite_Add (suite, "/bson/json/date", test_bson_json_date);
    TestSuite_Add (suite, "/bson/json/read/missing_complex", test_bson_json_read_missing_complex);
+   TestSuite_Add (suite, "/bson/json/read/invalid_binary", test_bson_json_read_invalid_binary);
    TestSuite_Add (suite, "/bson/json/read/invalid_json", test_bson_json_read_invalid_json);
    TestSuite_Add (suite, "/bson/json/read/bad_cb", test_bson_json_read_bad_cb);
    TestSuite_Add (suite, "/bson/json/read/invalid", test_bson_json_read_invalid);
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   TestSuite_Add (suite, "/bson/json/read/decimal128", test_bson_json_read_decimal128);
+#endif
    TestSuite_Add (suite, "/bson/json/read/file", test_json_reader_new_from_file);
    TestSuite_Add (suite, "/bson/json/read/bad_path", test_json_reader_new_from_bad_path);
    TestSuite_Add (suite, "/bson/json/read/invalid", test_bson_json_read_invalid);
-   TestSuite_Add (suite, "/bson/json/read/invalid", test_bson_json_read_invalid);
    TestSuite_Add (suite, "/bson/json/read/$numberLong", test_bson_json_number_long);
+   TestSuite_Add (suite, "/bson/json/read/$numberLong/zero", test_bson_json_number_long_zero);
    TestSuite_Add (suite, "/bson/json/read/dbref", test_bson_json_dbref);
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   TestSuite_Add (suite, "/bson/as_json/decimal128", test_bson_as_json_decimal128);
+   TestSuite_Add (suite, "/bson/json/read/$numberDecimal", test_bson_json_number_decimal);
+#endif
 }
